from dotenv import load_dotenv
from os import environ
from os import makedirs
from pytgpt.utils import api_static_dir

load_dotenv()

assert environ.get("token"), "Export bot's token to the environment"

bot_token: str = environ.get("token")
admin_id: int = int(environ.get("admin-id", 1234567))
max_tokens: int = int(environ.get("max-tokens", 600))
timeout: int = int(environ.get("timeout", 30))
voice: str = environ.get("voice", "Brian")
loglevel: int = int(environ.get("loglevel", 20))
logfile = environ.get("logfile")

bot_dir = api_static_dir / "bots"
path_to_db = bot_dir / "telegram.db"

makedirs(bot_dir, exist_ok=True)



import telebot
import pytgpt.auto as text_generator
import pytgpt.imager as image_generator
from pytgpt.utils import Audio as audio_generator
from pytgpt.utils import Conversation
from pytgpt.utils import AwesomePrompts
from functools import wraps
import logging

from .config import bot_token, max_tokens, timeout, voice, loglevel, logfile, admin_id
from .db import User, Chat

log_params = dict(
    format="%(asctime)s : %(levelname)s - %(message)s",
    datefmt="%d-%b-%Y %H:%M:%S",
    level=loglevel,
)

if logfile:
    log_params["filename"] = logfile

logging.basicConfig(**log_params)

bot = telebot.TeleBot(bot_token, parse_mode="Markdown")

bot.remove_webhook()

logging.info(f"Bot started sucessfully. Admin ID - [{admin_id}]")


def handler_formatter(func):

    @wraps(func)
    def decorator(message: telebot.types.Message):
        try:
            logging.info(
                f"Serving user [{message.from_user.id}] ({message.from_user.full_name}) - Function [{func.__name__}]"
            )
            if message.text and message.text.startswith("/"):
                message.text = " ".join(message.text.split(" ")[1:])
            return func(message)
        except Exception as e:
            logging.error(f"Error on function - {func.__name__} - {e}")
            return bot.reply_to(
                message, text="An error occured and could't complete that request."
            )

    return decorator


@bot.message_handler(commands=["help", "start"])
@handler_formatter
def home(message: telebot.types.Message):
    """
    Welcome to [pytgpt-bot](https://github.com/Simatwa/pytgpt-bot).
    /help : Show this help info.
    /chat : Chat with AI.
    /imager : Generate image from text. (default)
    /prodia : Generate image from text. (Prodia)
    /audio : Generate audio from text.
    /intro : Check current chat intro.
    /history : Check chat history.
    /sintro : Set new value for chat intro.
    /reset : Start new chat thread.
    /myid : Echo your Telegram ID.
    /default : Chat with AI.
    """

    return bot.send_message(
        message.chat.id,
        text=home.__doc__,
    )


@bot.message_handler(commands=["myid"])
@handler_formatter
def echo_user_id(message: telebot.types.Message):
    return bot.reply_to(
        message,
        f"Greetings {message.from_user.first_name}. Your Telegram ID is {message.from_user.id}.",
    )


@bot.message_handler(commands=["sintro"])
@handler_formatter
def set_chat_intro(message: telebot.types.Message):
    """Set new value for chat intro"""
    intro = AwesomePrompts().get_act(message.text) or message.text
    if not len(intro) > 10:
        return bot.reply_to(
            message, "The chat introduction must be at least 10 characters long."
        )
    user = User(message.from_user.id)
    user.update_intro(intro)
    bot.reply_to(message, "New intro set successfully.")


@bot.message_handler(commands=["intro"])
@handler_formatter
def check_chat_intro(message: telebot.types.Message):
    user = User(message.from_user.id)
    return bot.reply_to(message, user.chat_intro)


@bot.message_handler(commands=["history"])
@handler_formatter
def check_chat_history(message: telebot.types.Message):
    user = User(message.from_user.id)
    return bot.reply_to(message, user.chat_history or "Your chat history is empty.")


@bot.message_handler(commands=["imager", "img"])
@handler_formatter
def text_to_image_default(message: telebot.types.Message):
    """Generate image using `imager`"""
    if not message.text:
        return bot.reply_to(message, f"Text is required.")
    generator_obj = image_generator.Imager(
        timeout=30,
    )
    return bot.send_photo(
        message.chat.id,
        photo=generator_obj.generate(
            message.text,
        )[0],
        caption=message.text,
    )


@bot.message_handler(commands=["prodia", "prod"])
@handler_formatter
def text_to_image_prodia(message: telebot.types.Message):
    """Generate image using `prodia`"""
    if not message.text:
        return bot.reply_to(message, f"Text is required.")
    generator_obj = image_generator.Prodia(timeout=timeout)
    return bot.send_photo(
        message.chat.id,
        photo=generator_obj.generate(message.text)[0],
        caption=message.text,
    )


@bot.message_handler(commands=["audio", "aud"])
@handler_formatter
def text_to_audio(message: telebot.types.Message):
    """Convert text to audio"""
    if not message.text:
        return bot.reply_to(message, f"Text is required.")
    audio_chunk = audio_generator.text_to_audio(
        message=message.text,
        voice=voice,
        timeout=timeout,
    )
    return bot.send_audio(message.chat.id, audio=audio_chunk, caption=message.text)


@bot.message_handler(commands=["reset"])
@handler_formatter
def reset_chat(message: telebot.types.Message):
    """Reset current chat thread"""
    user = User(message.from_user.id)
    user.delete()
    return bot.reply_to(
        message,
        "New chat instance created.",
    )


@bot.message_handler(commands=["clear_database"])
@handler_formatter
def clear_chats(message: telebot.types.Message):
    """Delete all chat entries"""
    user = User(message.from_user.id)
    if not user.is_admin:
        return bot.reply_to(message, "Action resticted to admin only!")
    else:
        Chat.query("DELETE FROM Chat")
        logging.warning(
            f"Clearing Chats - [{message.from_user.full_name}] ({message.from_user.id}, {message.from_user.username})"
        )
        return bot.reply_to(message, "Chats cleared successfully.")


@bot.message_handler(content_types=["text"])
@handler_formatter
def text_chat(message: telebot.types.Message):
    """Text generation"""
    user = User(message.from_user.id)
    chat_record = user.record
    conversation = Conversation(max_tokens=max_tokens)
    conversation.chat_history = chat_record.get("history") or ""
    conversation_prompt = conversation.gen_complete_prompt(
        message.text, intro=chat_record.get("intro")
    )
    bot.send_chat_action(message.chat.id, "typing")
    ai_response = text_generator.AUTO(is_conversation=False, timeout=timeout).chat(
        conversation_prompt
    )
    conversation.update_chat_history(
        prompt=message.text, response=ai_response, force=True
    )
    user.update_history(conversation.chat_history)
    return bot.send_message(message.chat.id, text=ai_response)


@bot.message_handler(func=lambda val: True)
def any_other_action(message):
    return bot.reply_to(message, home.__doc__)



import sqlite3
from .config import admin_id
from .config import path_to_db
from functools import wraps
import logging
import typing
from pytgpt.utils import Conversation


def exception_handler(func):

    @wraps(func)
    def decorator(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            logging.error("Error on function" f" '{func.__name__}' - " f"{e}")

    return decorator


class Chat:
    """Performs CRUD operations on dbs"""

    @staticmethod
    @exception_handler
    def initialize():
        """Creates database"""
        conn = sqlite3.connect(path_to_db)
        conn.execute(
            f"""
        CREATE TABLE IF NOT EXISTS Chat(
            Id INTEGER PRIMARY KEY,
            Intro TEXT DEFAULT "{Conversation.intro}",
            History TEXT);
        """
        )
        conn.commit()

    @staticmethod
    @exception_handler
    def insert(**kwargs):
        """Inserts data
        Kwargs:  {
           field : data
        }
        """
        fields: list[str] = list(kwargs.keys())
        values: tuple[str] = tuple(kwargs.values())
        conn = sqlite3.connect(path_to_db)
        csr = conn.cursor()
        sql = (
            "INSERT INTO Chat"
            f"({', '.join(fields)}) "
            f"VALUES ({ '?,'*(len(values)-1) + '?' if len(values) > 1 else '?'})"
        )
        csr.execute(sql, values)
        csr.close()
        conn.commit()
        conn.close()

    @staticmethod
    @exception_handler
    def update(id, field, data):
        """Update record
        Args:
           id : User id
           field : Column name.
           data : Data value.
        """
        conn = sqlite3.connect(path_to_db)
        csr = conn.cursor()
        csr.execute(f"UPDATE  Chat SET  {field} = ?  WHERE id = ?" "", (data, id))
        csr.close()
        conn.commit()
        conn.close()

    @staticmethod
    @exception_handler
    def read(id: int, fetchone: bool = False) -> list[tuple]:
        """Read user data

        Args:
            id (int): User id
            fetchone (bool): Return first entry only. Defaults to False.

        Returns:
            tuple[tuple]: User data
        """
        conn = sqlite3.connect(path_to_db)
        csr = conn.cursor()
        csr.execute("SELECT * FROM Chat WHERE Id = ?", (id,))
        all_entries = csr.fetchone() if fetchone else csr.fetchall()
        csr.close()
        conn.close()
        return all_entries

    @staticmethod
    @exception_handler
    def delete(id: int):
        """Delete chat

        Args:
            id (int): User id

        Returns:
            None: None
        """
        conn = sqlite3.connect(path_to_db)
        conn.execute(f"DELETE FROM Chat WHERE ID={id}")
        conn.commit()
        conn.close()

    @staticmethod
    @exception_handler
    def query(sql: str) -> typing.Union[tuple[tuple], typing.NoReturn]:
        """Run sql commands and return entries"""
        conn = sqlite3.connect(path_to_db)
        csr = conn.cursor()
        csr.execute(sql)
        entries = csr.fetchall()
        csr.close()
        conn.close()
        return entries


class User(Chat):
    """User chat model"""

    def __init__(self, id: int):
        """Constructor

        Args:
            id (int): User id
        """
        self.id = id
        if Chat.read(self.id, True) is None:
            Chat.insert(id=self.id)

    @property
    def is_admin(self) -> bool:
        """Checks user admin status"""
        return self.id == admin_id

    @property
    def record(self) -> typing.Dict[str, typing.Union[int, str]]:
        """User data

        Returns:
            typing.Dict[str, typing.Union[int, str]]: Records
        """
        id, intro, history = Chat.read(id=self.id, fetchone=True)
        return dict(id=id, intro=intro, history=history)

    @property
    def chat_history(self) -> typing.Union[str, None]:
        """User's chat history"""
        return self.record.get("history")

    @property
    def chat_intro(self) -> str:
        """User's chat intro"""
        return self.record.get("intro")

    def update_history(self, data: str):
        """Update chat history

        Args:
            data (str): New history
        """
        Chat.update(
            self.id,
            field="history",
            data=data,
        )

    def update_intro(self, data: str):
        """Update chat intro

        Args:
            data (str): New intro.
        """
        Chat.update(
            self.id,
            field="intro",
            data=data,
        )

    def delete(self):
        """Delete current user account"""
        Chat.delete(self.id)


# import os
# os.remove(path_to_db)
Chat.initialize()
# exit()
# resp = User(113)
# resp.update_intro('hello world')
# resp.chat_history('hello')
# resp.delete()
# resp = User(113)
# print(resp.chat_intro)
